#include <unistd.h>

#include "test.h"
#include "../src/expression.h"
#include "../src/kmplayerplaylist.h"

int KMPlayer::connection_counter;

using namespace KMPlayer;

int main() {
    g_thread_init (NULL);
    g_type_init();
    String book_src = "<books>"
        "  <total>3</total>"
        "  <book price='10p'>"
        "    <title>Tom Sawyer</title>"
        "    <author>Mark Twain</author>"
        "  </book>"
        "  <book>"
        "    <title>Uncle Tom's Cabin</title>"
        "    <author>Harriet Beecher Stowe</author>"
        "  </book>"
        "  <book price='5$'>"
        "    <title>Pygmalion</title>"
        "    <author>George Bernard Shaw</author>"
        "  </book>"
        "</books>";
    TextStream book_in (&book_src);
    NodePtr books = new Document (String ());
    readXML (books, book_in, String ());
    {
        TEST_INFO (Test reading books);
        TEST_VERIFY (books->childNodes ().length () == 1);
    }
    {
        TEST_INFO (Test /books/book == 3);
        Expression *expr = evaluateExpr ("/books/book");
        expr->setRoot (books);
        TEST_VERIFY (expr->toInt () == 3);
        delete expr;
    }
    {
        TEST_INFO (Test count(/books/book) == 3);
        Expression *expr = evaluateExpr ("count(/books/book)");
        expr->setRoot (books);
        TEST_VERIFY (expr->toInt () == 3);
        delete expr;
    }
    {
        TEST_INFO (Test "//book == 3");
        Expression *expr = evaluateExpr ("//book");
        expr->setRoot (books);
        TEST_VERIFY (expr->toInt () == 3);
        delete expr;
    }
    {
        TEST_INFO (Test "/books/book[@price='5$'] == 1");
        Expression *expr = evaluateExpr ("/books/book[@price='5$']");
        expr->setRoot (books);
        Sequence *lst = expr->toSequence ();
        TEST_VERIFY (lst->length () == 1);
        delete lst;
        delete expr;
    }
    {
        TEST_INFO (Test "/books/book[@price][2]/title == 'Pygmalion'");
        Expression *expr = evaluateExpr ("/books/book[@price][2]/title");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Pygmalion");
        delete expr;
    }
    {
        TEST_INFO (Test "/books/book[@price][2][]/title == 'Pygmalion'");
        Expression *expr = evaluateExpr ("/books/book[@price][2][]/title");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Pygmalion");
        delete expr;
    }
    {
        TEST_INFO (Test "@price of third book] == '5$'");
        Expression *expr = evaluateExpr ("/books/book[3]/@price");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "5$");
        delete expr;
    }
    {
        TEST_INFO (Test "/books/book[@price='5$']/title == Pygmalion");
        Expression *expr = evaluateExpr ("/books/book[@price='5$']/title");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Pygmalion");
        delete expr;
    }
    {
        TEST_INFO (Test "/hook/book[3]/@price == '5$'");
        Expression *expr = evaluateExpr ("/books/book[3]/@price");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "5$");
        delete expr;
    }
    {
        TEST_INFO (Test "@price == '5$' using last book as root");
        Expression *expr = evaluateExpr ("/books/book[last()]");
        expr->setRoot (books);
        Sequence *lst = expr->toSequence ();
        TEST_VERIFY (lst->first ());
        if (lst->first ()) {
            NodePtr b3 = lst->first ()->data.node;
            Expression *expr2 = evaluateExpr ("@price");
            expr2->setRoot (b3);
            TEST_VERIFY (expr2->toString () == "5$");
            delete expr2;
        }
        delete lst;
        delete expr;
    }
    {
        TEST_INFO (Test sorting "/hook/book" on /title);
        Expression *expr = evaluateExpr ("sort('/books/book','/title')");
        expr->setRoot (books);
        Sequence *lst = expr->toSequence ();
        NodeValueItem *itm = lst->first ();
        TEST_VERIFY (itm);
        if (itm) {
            TEST_VERIFY (itm->data.node->firstChild ()->nodeValue () ==
                    "Pygmalion");
            itm = itm->nextSibling ();
            TEST_VERIFY (itm);
            if (itm) {
                TEST_VERIFY (itm->data.node->firstChild ()->nodeValue () ==
                        "Tom Sawyer");
                itm = itm->nextSibling ();
                TEST_VERIFY (itm);
                if (itm) {
                    TEST_VERIFY (itm->data.node->firstChild ()->nodeValue () ==
                            "Uncle Tom's Cabin");
                }
            }
        }
        delete lst;
        delete expr;
    }
    {
        TEST_INFO (Test sorting "/hook/book" on @price);
        Expression *expr = evaluateExpr ("sort('/books/book','@price')");
        expr->setRoot (books);
        Sequence *lst = expr->toSequence ();
        NodeValueItem *itm = lst->first ();
        TEST_VERIFY (itm);
        if (itm) {
            TEST_VERIFY (static_cast <Element *>(itm->data.node)->
                    getAttribute ("price").isEmpty ());
            itm = itm->nextSibling ();
            TEST_VERIFY (itm);
            if (itm) {
                TEST_VERIFY (static_cast <Element *>(itm->data.node)->
                        getAttribute ("price") == "10p");
                itm = itm->nextSibling ();
                TEST_VERIFY (itm);
                if (itm) {
                    TEST_VERIFY (static_cast <Element *>(itm->data.node)->
                            getAttribute ("price") == "5$");
                }
            }
        }
        delete lst;
        delete expr;
    }
    {
        TEST_INFO (Test "/books/book[@price='5$'] | //book[@price='10p'] => 2");
        Expression *expr = evaluateExpr ("/books/book[@price='5$'] | //book[@price='10p']");
        expr->setRoot (books);
        Sequence *lst = expr->toSequence ();
        TEST_VERIFY (lst->length () == 2);
        NodeValueItem *itm = lst->first ();
        if (itm) {
            TEST_VERIFY (static_cast <Element *>(itm->data.node)->
                    getAttribute ("price") == "5$");
            itm = itm->nextSibling ();
            if (itm)
                TEST_VERIFY (static_cast <Element *>(itm->data.node)->
                        getAttribute ("price") == "10p");
        }
        delete lst;
        delete expr;
    }
    {
        TEST_INFO (Test "count(/books/book[@price='5$'] | //book[@price='10p']) => 2");
        Expression *expr = evaluateExpr ("count(/books/book[@price='5$'] | //book[@price='10p'])");
        expr->setRoot (books);
        TEST_VERIFY (expr->toInt () == 2);
        delete expr;
    }
    {
        TEST_INFO (Test "string-join(//book[3]/title | //book[3]/@price, ' costs ')");
        Expression *expr = evaluateExpr ("string-join(//book[3]/title | //book[3]/@price, ' costs ')");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Pygmalion costs 5$");
        delete expr;
    }
    {
        TEST_INFO (Test "//book[not(@price)]/title == Uncle Tom's Cabin");
        Expression *expr = evaluateExpr ("//book[not(@price)]/title");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Uncle Tom's Cabin");
        delete expr;
    }
    {
        TEST_INFO (Test "escape-uri(//book[not(@price)]/title) == Uncle Tom's Cabin");
        Expression *expr = evaluateExpr ("escape-uri(//book[not(@price)]/title)");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Uncle%20Tom%27s%20Cabin");
        delete expr;
    }
    {
        TEST_INFO (Test "contains(//book[title=\"Uncle Tom's Cabin\"], 'Tom')");
        Expression *expr = evaluateExpr ("contains(//book[title=\"Uncle Tom's Cabin\"], 'Tom')");
        expr->setRoot (books);
        TEST_VERIFY (expr->toBool ());
        delete expr;
    }
    {
        TEST_INFO (Test "//book/title[contains(.,'Tom')]");
        Expression *expr = evaluateExpr ("//book/title[contains(.,'Tom')]");
        expr->setRoot (books);
        Sequence *lst = expr->toSequence ();
        TEST_VERIFY (lst->length () == 2);
        NodeValueItem *itm = lst->first ();
        if (itm) {
            TEST_VERIFY (itm->data.value () == "Tom Sawyer");
            itm = itm->nextSibling ();
            if (itm)
                TEST_VERIFY (itm->data.value () == "Uncle Tom's Cabin");
        }
        delete lst;
        delete expr;
    }
    {
        TEST_INFO (Test /books/book[number(/books/total)]/title == 'Pygmalion');
        Expression *expr = evaluateExpr ("/books/book[number(/books/total)]/title");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Pygmalion");
        delete expr;
    }
    {
        TEST_INFO (Test /books/book[author='George Bernard Shaw']/title == 'Pygmalion');
        Expression *expr = evaluateExpr ("/books/book[author='George Bernard Shaw']/title");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Pygmalion");
        delete expr;
    }
    {
        TEST_INFO (Test substring-after(/books/book[2]/title, 'Tom') == 's Cabin');
        Expression *expr = evaluateExpr ("substring-after(/books/book[2]/title, 'Tom')");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "'s Cabin");
        delete expr;
    }
    {
        TEST_INFO (Test substring-before(/books/book[2]/title, 'Tom') == 'Uncle ');
        Expression *expr = evaluateExpr ("substring-before(/books/book[2]/title, 'Tom')");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Uncle ");
        delete expr;
    }
    {
        TEST_INFO (Test starts-with(/books/book[2]/title, 'Uncle') == true);
        Expression *expr = evaluateExpr ("starts-with(/books/book[2]/title, 'Uncle')");
        expr->setRoot (books);
        TEST_VERIFY (expr->toBool ());
        delete expr;
    }
    {
        TEST_INFO (Test /books/book[starts-with(., 'Uncle')]/author == 'Harriet Beecher Stowe');
        Expression *expr = evaluateExpr ("/books/book[starts-with(., 'Uncle')]/author");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Harriet Beecher Stowe");
        delete expr;
    }
    {
        TEST_INFO (Test /books/book[starts-with('Uncle')]/author == 'Harriet Beecher Stowe');
        Expression *expr = evaluateExpr ("/books/book[starts-with('Uncle')]/author");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Harriet Beecher Stowe");
        delete expr;
    }
    {
        TEST_INFO (Test string-length(/books/book[1]/title) == 10);
        Expression *expr = evaluateExpr ("string-length(/books/book[1]/title)");
        expr->setRoot (books);
        TEST_VERIFY (expr->toInt () == 10);
        delete expr;
    }
    {
        TEST_INFO (Test tokenize(/books/book[2]/title,'\\S+') == 3);
        Expression *expr = evaluateExpr ("tokenize(/books/book[2]/title,'\\S+')");
        expr->setRoot (books);
        TEST_VERIFY (expr->toInt () == 3);
        delete expr;
    }
    {
        TEST_INFO (Test tokenize(/books/book[2]/title,'\\S+')[1] == Uncle);
        Expression *expr = evaluateExpr ("tokenize(/books/book[2]/title,'\\S+')[1]");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Uncle");
        delete expr;
    }
    {
        TEST_INFO (Test tokenize(/books/book[2]/title,'\\S+')[position() = 1] == Uncle);
        Expression *expr = evaluateExpr ("tokenize(/books/book[2]/title,'\\S+')[position() = 1]");
        expr->setRoot (books);
        debugLog() << expr->toString() << endl;
        TEST_VERIFY (expr->toString () == "Uncle");
        delete expr;
    }
    {
        TEST_INFO (Test subsequence(tokenize(/books/book[2]/title,'\\S+'),1,1) == Uncle);
        Expression *expr = evaluateExpr ("subsequence(tokenize(/books/book[2]/title,'\\S+'),1,1)");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Uncle");
        delete expr;
    }
    {
        TEST_INFO (Test count(subsequence(tokenize(/books/book[2]/title,'\\S+'),3,0)) == '');
        Expression *expr = evaluateExpr ("count(subsequence(tokenize(/books/book[2]/title,'\\S+'),3,0))");
        expr->setRoot (books);
        TEST_VERIFY (expr->toInt () == 0);
        delete expr;
    }
    {
        TEST_INFO (Test subsequence(tokenize(/books/book[2]/title,'\\S+'),3,-1) == Uncle);
        Expression *expr = evaluateExpr ("subsequence(tokenize(/books/book[2]/title,'\\S+'),3,-1)");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Cabin");
        delete expr;
    }
    {
        TEST_INFO (Test string-join(subsequence(tokenize(/books/book[2]/title,'\\S+'),2),',') == 'Toms,Cabin');
        Expression *expr = evaluateExpr ("string-join(subsequence(tokenize(/books/book[2]/title,'\\S+'),2),',')");
        expr->setRoot (books);
        TEST_VERIFY (expr->toString () == "Tom's,Cabin");
        delete expr;
    }
    books->document ()->dispose ();
    return 0;
}
